package se.natusoft.tools.json

import groovy.transform.CompileStatic
import se.natusoft.lic.annotation.Apache_Software_License_2_0
import se.natusoft.lic.annotation.SourceAvailableAt

import java.nio.charset.StandardCharsets

@Apache_Software_License_2_0
@SourceAvailableAt( "https://github.com/tombensve/NS-Toolbox/" )

/**
 * Reads JSON data.
 *
 * Do note:
 * - JSON objects are treated as Map<String, Object>!
 * - JSON string types will be Java String instances.
 * - Numbers will be "int" or "long":s for non decimals.
 * - Decimals will be double or BigDecimal depending on size of number.
 */
@CompileStatic
abstract class AbstractJSONReader extends InputStreamReader {

    protected PushbackReader parseReader

    /** Done parsing when true. */
    protected done = false

    //
    // Constructors
    //

    AbstractJSONReader( InputStream is ) {
        super( is, StandardCharsets.UTF_8 )
        // Unless I'm more or less blind JSON.org no longer mention that JSON files should
        // be in UTF-8! I'm ignoring that and force UTF-8!

        // Since I want to be flexible and be able to read from an InputStream this extends
        // InputStreamReader! But since I also need to be able to do push back I'm wrapping
        // the actual Reader with PushBackReader. This means that the "parseReader" instance
        // MUST be used for all reading rather than this class which is a reader!!
        //
        // I really don't like this, but see no other solution!
        this.parseReader = new PushbackReader( this )

        // Start parsing stream.
        parse()
    }

    //
    // Methods
    //

    /**
     * Parses JSON
     */
    protected void parse() {

        skipWhitespace()

        char c = readChar()

        switch ( c ) {
            case '{':
                parseObject()
                break

            case '[':
                parseList()
                break

            case '"':
                parseString()
                break

            default:
                throw new JSONException("Bad input! '${c}'")
        }
    }

    /**
     * Reads and returns a single character.
     * @throws JSONException on bad input.
     *
     * THIS SHOULD ALWAYS BE USED TO READ!!
     */
    protected char readChar() {
        char[] buffer = new char[1]

        if ( this.parseReader.read( buffer ) != 1 ) {
            throw new JSONException( "Bad Input!" )
        }

        return buffer[ 0 ]
    }

    /**
     * Reads characters up to specified character.
     *
     * @param until The character to read until.
     *
     * @return The characters between current point to until character.
     */
    protected String readUntil( char until ) {

        StringBuilder result = new StringBuilder()
        boolean done = false

        while ( !done ) {
            char current = readChar()

            if ( current == until ) {
                done = true
            } else {
                result.append( current )
            }
        }

        result
    }

    /**
     * Forwards to next non whitespace character.
     */
    protected void skipWhitespace() {

        char chr = readChar()

        while ( chr == ' ' as char || chr == '\n' as char || chr == '\t' as char || chr == '\r' as char ) {
            chr = readChar()
        }

        char[] unreadChar = new char[1]
        unreadChar[ 0 ] = chr
        this.parseReader.unread( unreadChar )
    }

    //
    // Parsing of different types.
    //

    /**
     * Parses an object.
     */
    void parseObject() {

    }

    /**
     * Parses a list.
     */
    void parseList() {

    }

    /**
     * Parses a string.
     */
    void parseString() {

    }

    /**
     * Parses a value. Have to handle both int:s and decimals, and possibly BigDecimal!
     */
    void parseValue() {

    }
}
